/*
 * stunnelw - Win32 User Interface for Stunnel
 *
 * Copyright (c) 2000 Robert Spier <rspier@pobox.com>
 *               All Rights Reserved
 *
 * see stunnel.c for more information about stunnel 
 *
 * Special Thanks for this file:
 *   Ron Isaacson
 *   Charles Petzold
 *   And, Microsoft who gave me Petzold's book for free.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
 */

#ifdef USE_WIN32UI
#include <windows.h>
#include <stdio.h>
#include <malloc.h>
#include "resource.h"

#include "common.h"

#define exit win32_my_exit

HICON hicon_stunnel;
HWND g_hwnd;
HWND g_hwndEdit;
HINSTANCE g_hinstance;
int stunnel_main(int argc, char* argv[]);

#define IDM_LOG 1

/* windows provides us with a single string for the command line.
   stunnel_main wants it split into argc/argv */
void start_stunnel_main( LPSTR szCmdLine ) {
  char *argvp[128];
  int i=1;
  
  char *p = strtok(szCmdLine," ");
  while( p ) {
    argvp[i++] = p;
    p = strtok(NULL," ");
  }
  argvp[0]="stunnel";

  stunnel_main(i,argvp);
}

/* Win32 Message Handling Procedure */
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, 
			  LPARAM lParam) {
  static int toggleShow = 0;
  static HWND hwndStunnelInfo;
  static OPENFILENAME ofn;
  static char logFileName[256];

  switch(message) {
  case WM_CREATE:
    g_hwnd = hwnd;
    {
      DWORD threadID;
      
      g_hwndEdit = CreateWindow( TEXT("edit"), NULL,
				 WS_CHILD | WS_VISIBLE | WS_HSCROLL | 
				 WS_VSCROLL | WS_BORDER | ES_LEFT | 
				 ES_MULTILINE | ES_READONLY | 
				 ES_AUTOHSCROLL | ES_AUTOVSCROLL,
				 0,0,0,0, hwnd, (HMENU) IDM_LOG,
				 ((LPCREATESTRUCT)lParam)->hInstance, NULL );

	hwndStunnelInfo = CreateWindow( TEXT("static"), TEXT( STUNNEL_INFO ),
				   WS_CHILD| WS_VISIBLE | SS_SUNKEN | SS_RIGHT,
				   0,0,0,0, hwnd, NULL,
				   ((LPCREATESTRUCT)lParam)->hInstance, NULL );
				   
	update_trayicon( status_string );
	CloseHandle(CreateThread(NULL, 0, 
				 (LPTHREAD_START_ROUTINE)start_stunnel_main,
				 (void *)(((LPCREATESTRUCT)lParam)->lpCreateParams), 
				 0, &threadID));

	logFileName[0] = 0;
	ofn.lStructSize = sizeof( OPENFILENAME);
	ofn.hwndOwner = hwnd;
	ofn.hInstance=NULL;
	ofn.lpstrFilter = NULL;
	ofn.lpstrCustomFilter = NULL;
	ofn.nMaxCustFilter = 0;
	ofn.nFilterIndex = 0;
	ofn.lpstrFile = logFileName;
	ofn.nMaxFile = 256;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.lpstrTitle= NULL;
	ofn.Flags = OFN_OVERWRITEPROMPT;
	ofn.nFileOffset=0;
	ofn.nFileExtension=0;
	ofn.lpstrDefExt = TEXT("log");
	ofn.lCustData = 0;
	ofn.lpfnHook = NULL;
	ofn.lpTemplateName = NULL;
    }
    return 0;

  case WM_SETFOCUS:
    SetFocus(g_hwndEdit);
    
  case WM_USER:			/* coming from the tray icon */
    if ( lParam == WM_LBUTTONDBLCLK ) {
      toggleShow = !toggleShow;
      if (toggleShow) {
	ShowWindow(hwnd, SW_SHOW);
	SetForegroundWindow( hwnd );
      }
      else
	ShowWindow(hwnd, SW_HIDE);
    }


    if ( lParam == WM_RBUTTONUP ) {
      POINT pt;
      HMENU hMenu = LoadMenu( g_hinstance, MAKEINTRESOURCE(IDR_TRAYMENU));
      GetCursorPos(&pt);
      //      SetForegroundWindow( hwnd );
      TrackPopupMenu( GetSubMenu(hMenu,0), TPM_LEFTALIGN | TPM_BOTTOMALIGN | TPM_LEFTBUTTON,
		      pt.x, pt.y, 0, hwnd, NULL);
      DestroyMenu(hMenu);
    }
    return 0;
  case WM_SIZE:
    /* we didn't set the size of the window on creation, so set it now */
    MoveWindow( g_hwndEdit, 0,0, LOWORD(lParam), HIWORD(lParam)-0, TRUE);
    return 0;

    case WM_COMMAND:
      switch (LOWORD (wParam)) {
      case IDM_LOG:
      /* Possible errors from Log Editbox */
	if (HIWORD(wParam) == EN_ERRSPACE ||
	    HIWORD(wParam) == EN_MAXTEXT)
	  MessageBox(g_hwnd, TEXT("Edit control out of space."),
		     "stunnel", MB_OK | MB_ICONSTOP);
	return 0;

      case ID_SHOW_LOGWINDOW:
	SendMessage(hwnd,WM_USER,0,WM_LBUTTONDBLCLK);
	return 0;

      case ID_FILE_EXIT:
	SendMessage(hwnd,WM_DESTROY,0,0);
	return 0;

      case ID_HELP_ABOUT:
	MessageBox(g_hwnd, 
		   STUNNEL_INFO,
		   "About stunnel...",
		   MB_OK | MB_ICONINFORMATION );
	return 0;

      case ID_FILE_SAVE:
	if (!GetSaveFileName(&ofn))
	  return 0;

	{
	  FILE *f;
	  SetCursor(LoadCursor(NULL,IDC_WAIT));
	  ShowCursor(TRUE);

	  f = fopen(logFileName,"w+");
	  if (f) {
	    int slength = SendMessage( g_hwndEdit, WM_GETTEXTLENGTH, 0,0 ) + 1;
	    char *string = malloc(slength + 2);

	    SendMessage( g_hwndEdit, WM_GETTEXT, slength, (int)string );	    
	    fprintf(f,"%s",string);
	    fclose(f);
	  } else {
	    MessageBox( hwnd, TEXT("Cannot save logfile..."),
			"Error", MB_ICONERROR | MB_OK );
	  }
	    ShowCursor(FALSE);
	    SetCursor(LoadCursor(NULL,IDC_ARROW));
	}

      } // switch
      break;

  case WM_DESTROY: 
    {
      NOTIFYICONDATA nid;
      HICON icon = LoadIcon( g_hinstance, MAKEINTRESOURCE( IDI_STUNNEL )); 
      nid.cbSize = sizeof (NOTIFYICONDATA);
      nid.hWnd = g_hwnd;
      nid.uID = 0;
      nid.hIcon = icon;
      nid.uFlags = NIF_MESSAGE | NIF_TIP | NIF_ICON;
      nid.uCallbackMessage = WM_USER;
      Shell_NotifyIcon( NIM_DELETE, &nid );
      PostQuitMessage (0);
      return 0;
    }
  } 

  return DefWindowProc( hwnd, message, wParam, lParam );
}

/* Win32 main(), initialization and such */

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
  
  static TCHAR szAppName[] = TEXT ("stunnel");
  HWND hwnd;
  MSG msg;
  WNDCLASS wndclass;

  wndclass.style = CS_HREDRAW | CS_VREDRAW;
  wndclass.lpfnWndProc = WndProc;
  wndclass.cbClsExtra = 0;
  wndclass.cbWndExtra = 0;
  wndclass.hInstance = hInstance;
  wndclass.hIcon = LoadIcon( hInstance, MAKEINTRESOURCE(IDI_STUNNEL));
  wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
  wndclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH);
  wndclass.lpszMenuName = TEXT("IDR_STUNNELMENU");
  wndclass.lpszClassName = szAppName;

  RegisterClass(&wndclass);

  g_hinstance = hInstance;

  strncpy(status_string,"stunnel: initalizing",80);

  hwnd = CreateWindow( szAppName, TEXT( STUNNEL_INFO ), 
		       WS_OVERLAPPEDWINDOW, 
		       CW_USEDEFAULT, CW_USEDEFAULT, 
		       CW_USEDEFAULT, CW_USEDEFAULT, 
		       NULL, NULL, hInstance, lpCmdLine);
  
  ShowWindow( hwnd, SW_HIDE);
  UpdateWindow(hwnd);

  /* Win32 Message Loop */
  while( GetMessage(&msg, NULL, 0,0) ) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  return msg.wParam;
}

/* update the tooltip on the tray icon */   
void update_trayicon(LPSTR szIconStr) {
  static DWORD nim = NIM_ADD;
  HICON icon = LoadIcon(g_hinstance, MAKEINTRESOURCE(IDI_STUNNEL)); 
  NOTIFYICONDATA nid;


  nid.cbSize = sizeof (NOTIFYICONDATA);
  nid.hWnd = g_hwnd;
  nid.uID = 0;
  nid.uFlags = NIF_MESSAGE | NIF_TIP | NIF_ICON;
  nid.uCallbackMessage = WM_USER;
  nid.hIcon = icon;
  lstrcpy (nid.szTip, szIconStr);
  Shell_NotifyIcon (nim, &nid);
  nim = NIM_MODIFY;
  DestroyIcon( icon );
}      

/* append a message to the log buffer */
void win32_log(char *text) {
    int slength = SendMessage( g_hwndEdit, WM_GETTEXTLENGTH, 0,0 ) + 1;
    char *string = malloc(slength + strlen(text) + 4);
    int offset = 0;
    if (slength > 20000)	/* if we're getting long, cut first 5k out */
      offset = 5000;
    SendMessage( g_hwndEdit, WM_GETTEXT, slength, (int)string );
    strcat( string, text );
    strcat( string, "\r\n" );
    SendMessage( g_hwndEdit, WM_SETTEXT, strlen(string)-offset, 
		 (int)string+offset);
    free(string);
}

/* exit the current thread and bring the log window to the front */
void win32_my_exit(int v) {
  fprintf(stderr,"exit %d",v);
  win32_log("");
  win32_log("stunnel has terminated.\r\n"
	    "close this log window to terminate the process");
  ShowWindow( g_hwnd, SW_SHOW);
  SetForegroundWindow( g_hwnd );
  ExitThread(v);
}

/* append multiple \n seperated lines to the log buffer */
void win32_log_unixlines(char *text) {

  char *textcopy;
  char *line;
  textcopy = malloc(strlen(text)+1);
  strcpy(textcopy,text);
  line = strtok(textcopy,"\n");
  while ( line ) {
    win32_log(line);
    line = strtok(NULL,"\n");
  }
  free(textcopy);
}

#endif
